# Code taken from https://github.com/mherrmann/gitignore_parser
# Modified for synecure's purposes
# MIT License

import collections
import os
import re

from os.path import dirname, abspath
from pathlib import Path

def parse_gitignore(lines, full_path="?"):
	rules = []
	for i, line in enumerate(lines):
		line = line.rstrip('\n')
		rule = rule_from_pattern(line, None, source=(full_path, i + 1))
		if rule:
			rules.append(rule)
	return lambda file_path: any(r.match(file_path) for r in rules)

def rule_from_pattern(pattern, base_path=None, source=None):
	"""
	Take a .gitignore match pattern, such as "*.py[cod]" or "**/*.bak",
	and return an IgnoreRule suitable for matching against files and
	directories. Patterns which do not match files, such as comments
	and blank lines, will return None.
	Because git allows for nested .gitignore files, a base_path value
	is required for correct behavior. The base path should be absolute.
	"""
	if base_path and base_path != abspath(base_path):
		raise ValueError('base_path must be absolute')
	# Store the exact pattern for our repr and string functions
	orig_pattern = pattern
	# Early returns follow
	# Discard comments and seperators
	if pattern.strip() == '' or pattern[0] == '#':
		return
	# Discard anything with more than two consecutive asterisks
	if pattern.find('***') > -1:
		return
	# Strip leading bang before examining double asterisks
	if pattern[0] == '!':
		negation = True
		pattern = pattern[1:]
	else:
		negation = False
	# Discard anything with invalid double-asterisks -- they can appear
	# at the start or the end, or be surrounded by slashes
	for m in re.finditer(r'\*\*', pattern):
		start_index = m.start()
		if (start_index != 0 and start_index != len(pattern) - 2 and
				(pattern[start_index - 1] != '/' or
				 pattern[start_index + 2] != '/')):
			return

	# Special-casing '/', which doesn't match any files or directories
	if pattern.rstrip() == '/':
		return

	directory_only = pattern[-1] == '/'
	# A slash is a sign that we're tied to the base_path of our rule
	# set.
	anchored = '/' in pattern[:-1]
	if pattern[0] == '/':
		pattern = pattern[1:]
	if pattern[0] == '*' and pattern[1] == '*':
		pattern = pattern[2:]
		anchored = False
	if pattern[0] == '/':
		pattern = pattern[1:]
	if pattern[-1] == '/':
		pattern = pattern[:-1]
	regex = fnmatch_pathname_to_regex(
		pattern
	)
	if anchored:
		regex = ''.join(['^', regex])
	return IgnoreRule(
		pattern=orig_pattern,
		regex=regex,
		negation=negation,
		directory_only=directory_only,
		anchored=anchored,
		base_path=Path(base_path) if base_path else None,
		source=source
	)

whitespace_re = re.compile(r'(\\ )+$')

IGNORE_RULE_FIELDS = [
	'pattern', 'regex',  # Basic values
	'negation', 'directory_only', 'anchored',  # Behavior flags
	'base_path',  # Meaningful for gitignore-style behavior
	'source'  # (file, line) tuple for reporting
]


class IgnoreRule(collections.namedtuple('IgnoreRule_', IGNORE_RULE_FIELDS)):
	def __str__(self):
		return self.pattern

	def __repr__(self):
		return ''.join(['IgnoreRule(\'', self.pattern, '\')'])

	def match(self, abs_path):
		matched = False
		if self.base_path:
			rel_path = str(Path(abs_path).relative_to(self.base_path))
		else:
			rel_path = str(Path(abs_path))
		if rel_path.startswith('./'):
			rel_path = rel_path[2:]
		if re.search(self.regex, f"/{rel_path}/"):
			matched = True
		return matched


# Frustratingly, python's fnmatch doesn't provide the FNM_PATHNAME
# option that .gitignore's behavior depends on.
def fnmatch_pathname_to_regex(pattern):
	"""
	Implements fnmatch style-behavior, as though with FNM_PATHNAME flagged;
	the path seperator will not match shell-style '*' and '.' wildcards.
	"""
	i, n = 0, len(pattern)
	
	seps = [re.escape(os.sep)]
	if os.altsep is not None:
		seps.append(re.escape(os.altsep))
	seps_group = '[' + '|'.join(seps) + ']'
	nonsep = r'[^{}]'.format('|'.join(seps))

	res = []
	while i < n:
		c = pattern[i]
		i += 1
		if c == '*':
			try:
				if pattern[i] == '*':
					i += 1
					res.append('.*')
					if pattern[i] == '/':
						i += 1
						res.append(''.join([seps_group, '?']))
				else:
					res.append(''.join([nonsep, '*']))
			except IndexError:
				res.append(''.join([nonsep, '*']))
		elif c == '?':
			res.append(nonsep)
		elif c == '/':
			res.append(seps_group)
		elif c == '[':
			j = i
			if j < n and pattern[j] == '!':
				j += 1
			if j < n and pattern[j] == ']':
				j += 1
			while j < n and pattern[j] != ']':
				j += 1
			if j >= n:
				res.append('\\[')
			else:
				stuff = pattern[i:j].replace('\\', '\\\\')
				i = j + 1
				if stuff[0] == '!':
					stuff = ''.join(['^', stuff[1:]])
				elif stuff[0] == '^':
					stuff = ''.join('\\' + stuff)
				res.append('[{}]'.format(stuff))
		else:
			res.append(re.escape(c))
	res.insert(0, '(?ms)/')
	res.append('/')
	return ''.join(res)



import os
import sys
import json
import subprocess
import shlex
from coleo import Option, default, run_cli

from .utils import (
    get_config,
    get_config_path,
    edit_config,
    write_config,
    readlines,
    writelines,
    quote,
)
from .version import version as sy_version


def q(message=None):
    sys.exit(message)


def _realpath(filename):
    return os.path.realpath(os.path.expanduser(filename))


def _get_remote(cfg, name):
    if name in cfg:
        return cfg[name]

    return {
        "type": "ssh",
        "url": name,
        "paths": {os.getenv("HOME"): ""},
    }


def _cfg_from_url(name):
    return {
        "type": "ssh",
        "url": name,
        "paths": {os.getenv("HOME"): ""},
    }


def _check_remote(cfg, name, create=False, msg=None):
    if name not in cfg:
        if create:
            cfg[name] = _cfg_from_url(name)
        else:
            print(f"ERROR: remote '{name}' is not defined")
            if msg:
                print(msg)
            q()
    return cfg[name]


def _sort_paths(remote):
    return sorted(remote["paths"].items(), key=lambda kv: -len(kv[0]))


################
# Entry points #
################


def entry_bsync():
    from . import bsync


def entry_sy():
    run_cli(main)


def entry_sy_config():
    commands = {}
    for name, value in globals().items():
        parts = name.replace("_", "-").split("-", 1)
        if parts[0] == "config" and len(parts) > 1:
            curr = commands
            for part in parts[1:-1]:
                curr = curr.setdefault(part, {})
            curr[parts[-1]] = value
    run_cli(commands)


######
# sy #
######


def main():
    # List all sy directories
    # [alias: -l]
    list: Option & bool = default(False)

    # Version information
    version: Option & bool = default(False)
    if version:
        q(f"sy v{sy_version}")

    # Files to synchronize
    # [positional: *]
    files: Option

    # Name of the remote to sync with
    # [alias: -r]
    remote: Option = default(None)

    # Port to connect to
    # [alias: -p]
    port: Option = default(None)

    # Do a dry run
    # [alias: -n]
    dry_run: Option & bool = default(False)

    # List the commands sy will run
    show_plan: Option & bool = default(False)

    # Verbose output
    # [alias: -v]
    verbose: Option & bool = default(False)

    # Prompt for changes (necessary to resolve conflicts)
    # [alias: -i]
    interactive: Option & bool = default(False)

    # Resolve a conflict with "local" or "remote" copy (default: "prompt")
    resolve: Option = default("prompt")
    if resolve not in ("local", "remote", "prompt"):
        sys.exit("ERROR: resolve must be 'local', 'remote' or 'prompt'")

    # Resolve a conflict with local copy
    # [option: -1]
    # [false-options: -2]
    # [false-options-doc: Resolve a conflict with remote copy]
    resolve_local: Option & bool = default(None)
    if resolve_local is True:
        resolve = "local"
    if resolve_local is False:
        resolve = "remote"

    remotes = get_config("remotes.json")
    directories = get_config("directories.json")

    if list:
        for path, dest in directories.items():
            print(f"{path:50} {dest}")
        return

    commands = []
    if not files:
        files.append(".")

    for filename in files:
        filename = _realpath(filename)

        remote_name = _fill_remote(filename, remote, directories)
        remote_config = _get_remote(remotes, remote_name)
        remote_config["port"] = port

        commands += plan_sync(
            filename,
            remote,
            remote_config,
            dry=dry_run,
            verbose=verbose,
            interactive=interactive,
            resolve=resolve,
        )

    for command in commands:
        if verbose or show_plan:
            if isinstance(command, str):
                print(command)
            else:
                print(" ".join(map(shlex.quote, command)))
        if not show_plan:
            subprocess.run(command)

    write_config("directories.json", directories, silent=True)


def _fill_remote(path, remote_name, directories):
    if remote_name is None:
        if path not in directories:
            q("Please specify a destination")
        regdest = directories[path]
        assert regdest is not None
        return regdest
    else:
        directories[path] = remote_name
        return remote_name


def _check_dir(url, path, port):
    if url:
        args = ["ssh", url]
        if port:
            args.extend(["-p", str(port)])
        args.append(f"test -d {quote(path)}")
        result = subprocess.call(args)
        return result == 0

    else:
        return os.path.isdir(path)


def plan_sync(
    path,
    remote_name,
    remote,
    dry=False,
    verbose=False,
    interactive=False,
    resolve="prompt",
):

    for pfx, repl in _sort_paths(remote):
        if path.startswith(pfx):
            destpath = os.path.join(repl, path[len(pfx) + 1 :])
            break
    else:
        q(
            f"There is no rule to remap path '{path}' on '{remote_name}'"
            f"\nTry: 'sy-remote path {remote_name} <SRC_PREFIX> <DEST_PREFIX>'"
        )

    print(f"# SYNC LOCAL      {path}")
    if remote["type"] == "ssh":
        dest = f"{remote['url']}:{destpath}"
        print(f"# WITH REMOTE     {dest}")
    elif remote["type"] == "file":
        dest = destpath
        print(f"# WITH LOCAL      {dest}")
    else:
        print(f"# Unknown remote type: {remote['type']}")

    commands = []

    if os.path.exists(path):
        isdir = os.path.isdir(path)
    else:
        isdir = _check_dir(remote["url"], destpath, remote["port"])

    if isdir:
        # Use bsync to synchronize both directories

        cmdopts = ["-d"]
        if dry:
            cmdopts.append("-n")
        elif interactive:
            pass
        else:
            cmdopts.append("-y")

        if resolve == "local":
            cmdopts.append("-1")
        elif resolve == "remote":
            cmdopts.append("-2")

        if remote["type"] == "ssh" and remote["port"]:
            cmdopts.append(f"-p {remote['port']}")

        cmd = ["sy-bsync", *cmdopts, path, dest]
        commands.append(cmd)

    else:
        # Use two rsync commands to synchronize a single file
        # This will never erase the file
        destdir = os.path.dirname(destpath)

        common = ["rsync", "-ptu"]
        if verbose:
            common.append("-v")
        if remote["type"] == "ssh":
            common += [
                "-e",
                f"ssh ",
                f"-p {remote['port']}" if remote["port"] else "",
                # This is a dirty trick to create the directory on the remote
                "--rsync-path",
                f"mkdir -p {destdir}; rsync",
            ]
        elif remote["type"] == "file":
            commands.append(["mkdir", "-p", destdir])

        cmd1 = [*common, path, dest]
        cmd2 = [*common, dest, path]

        commands += [cmd1, cmd2]

    return commands


#############
# sy-remote #
#############


def config_add():
    """Add a new remote or edit an existing one."""
    # Name of the remote
    # [positional]
    name: Option

    # URL of the remote
    # [positional]
    url: Option

    if "://" not in url:
        q(
            "URL should be formatted as type://url -- currently accepted are:"
            "\n* SSH:              ssh://user@host"
            "\n* Local directory:  file:///path/from/root"
        )

    typ, url = url.split("://", 1)
    if typ not in ("ssh", "file"):
        q(f"Unknown protocol: '{typ}'. Accepted protocols are 'ssh' and 'file'.")

    if typ == "file":
        url = _realpath(url)

    cfg = get_config("remotes.json")
    if name in cfg:
        entry = cfg[name]
        entry["type"] = typ
        entry["url"] = url
    else:
        entry = {"type": typ, "url": url, "paths": {os.getenv("HOME"): ""}}
        cfg[name] = entry

    print(json.dumps(entry, indent=4))

    write_config("remotes.json", cfg)


def config_view():
    """View configuration for a remote."""
    # Name of the remote
    # [positional: ?]
    name: Option

    cfg = get_config("remotes.json")
    if name is not None:
        cfg = _check_remote(cfg, name)

    print(json.dumps(cfg, indent=4))


def config_list():
    """List existing remotes and paths."""
    cfg = get_config("remotes.json")
    for name, defn in cfg.items():
        print(f"{name:30} ({defn['type']}) {defn['url']}")
        for local_path, remote_path in defn["paths"].items():
            print(f"    {local_path:30} -> :{remote_path}")


def config_edit():
    """Edit the configuration file directly."""
    edit_config("remotes.json")


def config_ssh():
    """Edit your SSH configuration file."""
    edit_config(_realpath("~/.ssh/config"))


def _list_paths(remote):
    for s, d in _sort_paths(remote):
        print(f"{s:30}:{d}")


def config_add_path():
    """Add a path mapping for a remote."""
    # Name of the remote
    # [positional]
    name: Option

    # Source path
    # [positional]
    source: Option

    # Destination path
    # [positional]
    dest: Option

    cfg = get_config("remotes.json")
    remote = _check_remote(cfg, name, create=True)

    paths = remote["paths"]
    paths[source] = dest
    _list_paths(remote)
    write_config("remotes.json", cfg)


def config_list_paths():
    """List paths for a remote"""
    # Name of the remote
    # [positional]
    name: Option

    cfg = get_config("remotes.json")
    remote = _check_remote(cfg, name, create=False)
    _list_paths(remote)


def config_remove_path():
    """Remove a path mapping for a remote"""
    # Name of the remote
    # [positional]
    name: Option

    # Source path
    # [positional]
    source: Option

    cfg = get_config("remotes.json")
    remote = _check_remote(cfg, name, create=False)

    if source not in remote["paths"]:
        q(f"Source path '{source}' is not mapped")
    del remote["paths"][source]
    _list_paths(remote)
    write_config("remotes.json", cfg)


def config_remove():
    """Remove a remote."""
    # Name of the remote
    # [positional]
    name: Option

    cfg = get_config("remotes.json")
    _check_remote(cfg, name, create=False, msg="Nothing to remove")
    del cfg[name]
    write_config("remotes.json", cfg)


def config_ignore():
    """Add/remove global ignore patterns."""
    # Patterns to ignore
    # [positional: *]
    patterns: Option = default([])

    # List the ignores
    # [alias: -l]
    list: Option & bool = default(False)

    # Remove the ignores
    # [alias: -r]
    # [nargs: *]
    remove: Option = default([])

    ign = get_config_path("ignore")

    if list:
        print(open(ign).read(), end="")
        sys.exit(0)

    if not patterns:
        edit_config("ignore")
        sys.exit(0)

    lines = readlines(ign)
    new_lines = []

    for line in lines:
        if line in remove:
            print(f"-{line}")
        else:
            print(f" {line}")
            new_lines.append(line)

    no_add = {*lines, *remove}

    for pattern in patterns:
        if pattern not in no_add:
            print(f"+{pattern}")
            lines.append(pattern)

    writelines(ign, lines)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['synecure']

package_data = \
{'': ['*']}

install_requires = \
['coleo>=0.2.1,<0.3.0']

entry_points = \
{'console_scripts': ['sy = synecure.cli:entry_sy',
                     'sy-bsync = synecure.cli:entry_bsync',
                     'sy-config = synecure.cli:entry_sy_config']}

setup_kwargs = {
    'name': 'synecure',
    'version': '0.2.1',
    'description': 'File sync utility',
    'long_description': '\n# Synecure\n\nSynecure provides a command line program called `sy` that allows easy synchronization of files and directories over SSH between different machines. It is mostly a wrapper around [bsync](https://github.com/dooblem/bsync), which is itself based on the standard UNIX tool `rsync`.\n\nThis is beta software and comes with **NO GUARANTEES** that it will keep your data safe. It should not be used as backup solution.\n\n\n## Install\n\n```bash\npip install synecure\n```\n\n\n## Usage\n\n```bash\n# Sync local ~/directory with remote $HOME/directory on me@awesome.person\nsy ~/directory -r me@awesome.person\n\n# Sync current directory with the same path on me@awesome.person, port 2222\nsy -r me@awesome.person -p 2222\n\n# Synchronize to a named remote from your SSH config\nsy -r me\n\n# Synchronize the current directory to the last used remote (for that directory)\nsy\n```\n\nBy default, `sy` can take on any path within your `$HOME` and will set the corresponding path on the remote\'s `$HOME`. It is possible to change this behavior or synchronize paths outside of `$HOME` using the `sy-config add-path` command.\n\n`sy` with no argument will sync the current directory using the last remote for that directory (you will need to use the -r flag the first time, but not subsequently).\n\n\n## Howto\n\n\n### Ignore files\n\nAdd a `.bsync-ignore` file in the root directory to sync with a filename or glob pattern on each line, and they will be ignored. It works more or less like `.gitignore`.\n\nPutting `.bsync-ignore` files in subdirectories to ignore files in these subdirectories will unfortunately not work, so `sy ~/x` and `sy ~/x/y` may synchronize the contents of `~/x/y` differently if both directories contain different `.bsync-ignore` files, or if one has an ignore file and the other does not.\n\n\n### Global ignores\n\nThe `sy-config ignore` command can be used to generally ignore files or directories:\n\n```bash\n# Edit the ignore file using $EDITOR, if it is set\nsy-config ignore\n\n# List all existing ignores\nsy-config ignore -l\n\n# Ignore all files that end with ~\n# Do not forget the single quotes here, to avoid shell expansion!\nsy-config ignore \'*~\'\n\n# Unignore files that end with ~\nsy-config ignore -r \'*~\'\n```\n\nThe ignores work mostly like `.gitignore` or `.bsync-ignore` above, but they apply globally. Note that `sy` will also read *remote-side* global ignores when syncing to a remote. Global ignores are located at `$HOME/.config/synecure/ignore`, so a remote can define some global ignores even without installing `sy` remote-side. Global ignores local-side, remote-side, as well as `.bsync-ignore` files local-side and remote-side are all merged together.\n\n\n### Customize synchronization paths\n\nTo synchronize local `/etc` to remote `/etcetera`, for named remote `desktop`:\n\n```bash\nsy-config add-path desktop /etc /etcetera\n```\n\nObviously, this will only work if the remote user has the permissions to write to `/etcetera`. You can have multiple remotes for the same host with different users, if that helps.\n\nTo synchronize local `~/hello` to remote `~/bonjour`:\n\n```bash\nsy-config add-path desktop ~/hello bonjour\n```\n\nDon\'t use `~` for the remote path, it will complete to the wrong thing.\n\nTo list available remotes and paths:\n\n```bash\nsy-config list\n```\n\n### Sync local directories\n\n```bash\nsy-config add dropbox file://~/Dropbox\n```\n\n## Other options\n\n### Dry run\n\nUse the `-n` flag to perform a "dry run": `sy` (well, `bsync`) will report all the transfers that would occur but it will not perform them.\n\nUse `--show-plan` to get the sequence of commands that `sy` will run.\n\n### Conflict resolution\n\nWhenever a file was modified on both ends since the last sync, `sy` will ask which one you want to keep.\n\nUse `sy <options> --resolve local` (or `sy <options> -1`) to always keep the local file without prompting, or `--resolve remote` (or `-2`) to always keep the remote file.\n\n### List directories\n\n`sy -l` will list all directories that have been previously synced using the tool, along with the last remote they were synced to (remember that `sy` without the `-r` option will sync to the last remote).\n\n## Configuration files\n\n* `~/.config/synecure/remotes.json` defines protocols and paths for named remotes.\n  * You can open an editor for that file with `sy-config edit`\n* `~/.config/synecure/ignore` lists global ignores.\n  * You can open an editor for that file with `sy-config ignore`\n* `~/.config/synecure/directories.json` maps directories to last used remotes.\n* `~/.ssh/config` is the standard location to define host information for `ssh`.\n  * For convenience, you can open an editor for that file with `sy-config ssh`\n',
    'author': 'Olivier Breuleux',
    'author_email': 'breuleux@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/breuleux/synecure',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
